# -----------------------------------------------------------
# This script intersects city limits stored in a Shapefile
# with vectorized population data.
#City limits must be stored in a Mollweide-projected shapefile.
# Vincent Verbavatz - 2020
# -----------------------------------------------------------


import os
import matplotlib.pyplot as plt
from osgeo import gdal, ogr, osr
import sys

#folder where country shapefiles are stored
folder_countries = 'Cities/'
#folder where vectorized populations are stored
folder_population = 'Vectorized/'
#output folder for Intersected data
folder_result = 'Results/'

#Definition of Mollweide projection from Well known text
mollweide_wkt = 'PROJCS["World_Mollweide",GEOGCS["WGS 84",DATUM["WGS_1984", \
                SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]], \
                AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0],UNIT["Degree",0.0174532925199433]], \
                PROJECTION["Mollweide"],PARAMETER["central_meridian",0],PARAMETER["false_easting",0], \
                PARAMETER["false_northing",0],UNIT["metre",1,AUTHORITY["EPSG","9001"]], \
                AXIS["Easting",EAST],AXIS["Northing",NORTH]]'

mollweide = osr.SpatialReference()
mollweide.ImportFromWkt(mollweide_wkt)

#Use World Geodetic System 1984 to import stop coordinates
inSpatialRef = osr.SpatialReference()
inSpatialRef.ImportFromEPSG(4326)

#Defines Coordinate transformation
coordTrans = osr.CoordinateTransformation(inSpatialRef, mollweide)

densityField = ogr.FieldDefn('Density', ogr.OFTReal)

def density(layer):
    """
    Calculates density from population field.
    """
    if layer[0].GetFieldCount() == 1:
         layer.CreateField(densityField)

    for i in range(len(layer)):
        feature = layer[i]
        ingeom = feature.GetGeometryRef()
        area = ingeom.GetArea()
        feature.SetField('Density', feature.GetField('Population') / area)
        layer.SetFeature(feature)

def inverse_density(layer):
    """
    Calculates population from density field.
    """
    for i in range(len(layer)):
        feature = layer[i]
        ingeom = feature.GetGeometryRef()
        area = ingeom.GetArea()
        feature.SetField('Population', feature.GetField('Density') * area)
        layer.SetFeature(feature)

def remove_layer(name):
    """
    Removes a layer from system
    """
    try:
        os.remove(name + '.shp')
        os.remove(name + '.shx')
        os.remove(name + '.prj')
        os.remove(name + '.dbf')
    except IOError:
        pass

def createBuffer(inputlyr, bufferlyr, bufferDist):
    """
    From the Python GDAL/OGR Cookbook, licensed under the MIT license.
    This recipe buffers features of a layer and saves them to a new Layer.
    """
    featureDefn = bufferlyr.GetLayerDefn()

    for feature in inputlyr:
        ingeom = feature.GetGeometryRef()
        geomBuffer = ingeom.Buffer(bufferDist)

        outFeature = ogr.Feature(featureDefn)
        outFeature.SetGeometry(geomBuffer)
        outFeature.SetField('Population', feature.GetField('Population'))
        bufferlyr.CreateFeature(outFeature)

        outFeature = None

drv = ogr.GetDriverByName("ESRI Shapefile")

for file in os.listdir(folder_countries):
    if file.endswith(".shp"):
        src_ds = ogr.Open(folder_countries + file, update = 0)
        Layer_ref = src_ds.GetLayer()

        srs = mollweide
        popField = ogr.FieldDefn('Population', ogr.OFTInteger)

        for city in Layer_ref:
            #replaces invalid names for cities
            city_name = city.GetField('fuaname_en').replace('/', '_')

            if os.path.isfile(folder_result + city_name + '.shp') is False:

                dst_ds = drv.CreateDataSource(folder_result + "temp_" + city_name + '.shp')

                #ancillary copy of the city
                tmp_city_layer = dst_ds.CreateLayer('temp_' + city_name,
                geom_type=ogr.wkbPolygon, srs = srs)
                tmp_city_layer.CreateFeature(city)

                #pre_corrected output layer
                city_layer = dst_ds.CreateLayer('unvalid_' + city_name,
                geom_type=ogr.wkbPolygon, srs = srs)
                city_layer.CreateField(popField)
                #city_layer.CreateField(densityField)

                #final and corrected output layer
                valid_city_layer = dst_ds.CreateLayer(city_name,
                geom_type=ogr.wkbPolygon, srs = srs)
                valid_city_layer.CreateField(popField)
                #valid_city_layer.CreateField(densityField)

                #ancillary empty layer
                empty_layer = dst_ds.CreateLayer('empty', geom_type=ogr.wkbPolygon, srs = srs)
                empty_layer.CreateField(popField)
                #empty_layer.CreateField(densityField)

                #tries to intersect the city with all population files
                for pop_file in os.listdir(folder_population):
                    if pop_file.endswith(".shp"):

                        ds_pop = ogr.Open(folder_population + pop_file, update = 0)
                        pop_layer = ds_pop.GetLayer()

                        #Calculates density of pop_grid
                        #density(pop_layer)

                        #intersection of the city with the current population file
                        part_city_layer = dst_ds.CreateLayer(city_name + "_" + pop_layer.GetName(),
                        geom_type=ogr.wkbPolygon, srs = srs)
                        part_city_layer.CreateField(popField)
                        part_city_layer.CreateField(densityField)

                        pop_layer.Intersection(tmp_city_layer, part_city_layer)

                        #checks if city and population file intersect
                        if part_city_layer.GetFeatureCount() > 0:
                            part_union = dst_ds.CreateLayer("part_union_" +
                            city_name + "_" + pop_layer.GetName(),
                             geom_type=ogr.wkbPolygon, srs = srs)
                            part_union.CreateField(popField)

                            #copies city_layer to part_union
                            empty_layer.Intersection(city_layer, part_union)

                            #adds part of the city to the final file
                            part_city_layer.Union(part_union, city_layer)

                            remove_layer(folder_result + "part_union_" + city_name + "_" + pop_layer.GetName())

                        remove_layer(folder_result + city_name + "_" + pop_layer.GetName())

                #Corrects geometry errors
                createBuffer(city_layer, valid_city_layer, 0.0)

                #inverse_density(valid_city_layer)

                remove_layer(folder_result +'temp_' + city_name)
                remove_layer(folder_result +'unvalid_' + city_name)
                remove_layer(folder_result + "empty")
                dst_ds.Destroy()
