# -----------------------------------------------------------
# This script provides vectorization tools to convert raster
# data into polygonized elements.
# All adjacent pixels with the same population are joined in
# a common polygon of population the sum of pixel populations.
#
# Vincent Verbavatz - 2020
# -----------------------------------------------------------

import os
import matplotlib.pyplot as plt
from osgeo import gdal, ogr, osr
import sys

#folder where raster data is stored
continent = "CentralAmerica/"

def pop_to_density(feature, area):
    """ Converts Population value of a feature into associated density """
    return max(0, feature.GetField('Population') / area)

def geo_correction(inputlyr, outputlyr):
    """
    Corrects geometry errors resulting from vectorization and recalculates
    poygon populations.
    Output layer should point to an empty layer.
    """
    newField = ogr.FieldDefn('Population', ogr.OFTInteger)
    outputlyr.CreateField(newField)
    featureDefn = outputlyr.GetLayerDefn()

    for feature in inputlyr:
        #corrects geometry errors with buffer of size 0
        ingeom = feature.GetGeometryRef()
        geomBuffer = ingeom.Buffer(0.0)

        #recalculates polygon population
        area = ingeom.GetArea()
        density = pop_to_density(feature, 62500)
        outFeature = ogr.Feature(featureDefn)
        outFeature.SetGeometry(geomBuffer)
        outFeature.SetField('Population', density * area)
        outputlyr.CreateFeature(outFeature)

        outFeature = None

for file in os.listdir(continent):
    if file.endswith(".tif"):

        basename = os.path.basename(file)
        mask = None
        file = continent + basename

        #import raster data
        src_ds = gdal.Open(file)
        srcband = src_ds.GetRasterBand(1)
        srs = osr.SpatialReference()
        srs.ImportFromWkt(src_ds.GetProjection())
        dst_layername = basename[:-4]

        #Prepares layer for vectorized data
        drv = ogr.GetDriverByName("ESRI Shapefile")
        dst_ds = drv.CreateDataSource(dst_layername + ".shp")
        dst_layer = dst_ds.CreateLayer(dst_layername, srs = srs)
        newField = ogr.FieldDefn('Population', ogr.OFTInteger)
        dst_layer.CreateField(newField)

        gdal.Polygonize(srcband, None, dst_layer, 0, [], callback=None)

        #Prepares corrected layer
        valid_dst_layer = dst_ds.CreateLayer("valid_" + dst_layername,
        geom_type=ogr.wkbPolygon, srs = srs)
        geo_correction(dst_layer, valid_dst_layer)

        dst_ds.Destroy()
        drv.DeleteDataSource(dst_layername + ".shp")
